//======================================================== IBJmth.c
//
// Mathematical Routines for IBJ-programs
// ======================================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-07-31  lj
// 2000-07-27 uj  extended for IBJodor
//
//==================================================================

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "IBJmsg.h"
#include "IBJmth.h"

char *MthVersion = "1.0.a";

float Pi = 3.14158265359;
float Ga[5] = { -0.90618, -0.53847, 0, 0.53847, 0.90618 };
float Gg[5] = { 0.236925, 0.47863, 0.56889, 0.47863, 0.236925 };

static char *eMODn = "IBJmth";

//============================================================ Erfc
double Erfc(       // Complementary Error Function
double x )
  {
  double a, t, r, s, f;
  static double rwpi = 0.564189583;
  int i;
  a = (x < 0) ? -x : x;
  if (a < 2.7577) {
    t = 1/(1 + 0.3275911*a);
    r = t*(  0.254829592 +
        t*( -0.284496736 +
        t*(  1.421413741 +
        t*( -1.453152027 +
        t*   1.061405429 )))); }
  else {
    s = 1.0;
    f = 1/(2*a*a);
    t = 1.0;
    for (i=0; i<6; i++) { t *= -(2*i+1)*f;  s += t; }
    r = s*rwpi/a; }
  if (a < 20) r *= exp(-a*a);
  else  r = 0.0;
  if (x < 0.0)  r = 2.0 - r;
  return r; }

//============================================================= Gamma
double Gamma( double x )
  {
  double f, g, z;
  static double b[9] = { 1.0, -0.577191652, 0.988205891, -0.897056937,
    0.918206857, -0.756704078, 0.482199394, -0.193527818, 0.035868343 };
  static double w2pi = 2.506628274631;
  int i;
  f = 1.0;
  if (x <= 0.0)  return 0;
  if (x > 10.0) {  /* Stirling */
    z = 1/x;
    f = 1+z*(1.0/12.0+z*(1.0/288.0+z*(-139.0/51840-z*571.0/2488320.0)));
    g = exp(-x)*pow(x,x-0.5)*w2pi;
    return f*g;
    }
  if (x < 1.0) f /= x;
  else {
    while (x > 2.0) {
      x -= 1.0;
      f *= x;
      }
    x -= 1.0;
    }
  g = b[8];
  for (i=7; i>=0; i--)  g = b[i] + x*g;
  return f*g;
  }

double GammaP( double x, double p )
  {
  double f, g, z;
  static double w2pi = 2.506628274631;
  if (x < 10.0) {
    f = Gamma(x);
    z = pow(f, p);
    return z;
    }
  z = 1/x;
  f = w2pi*(1+z*(1.0/12.0+z*(1.0/288.0+z*(-139.0/51840-z*571.0/2488320.0))));
  g = exp(-p*x)*pow(x,p*(x-0.5));
  z = g*pow(f,p);
  return z;
  }

double BssI0( double x )
  {
  double ax, ans, y;
  if ((ax=fabs(x)) < 3.75) {
    y = x/3.75;
    y *= y;
    ans = 1.0+y*(3.5156229+y*(3.0899424+y*(1.2067492
            +y*(0.2659732+y*(0.360768e-1+y*0.45813e-2)))));
    }
  else {
    y = 3.75/ax;
    ans = (exp(ax)/sqrt(ax))*(0.39894228+y*(0.1328592e-1
            +y*(0.225319e-2+y*(-0.157565e-2+y*(0.916281e-2
            +y*(-0.2057706e-1+y*(0.2635537e-1+y*(-0.1647633e-1
            +y*0.392377e-2))))))));
    }
  return ans;
  }

/*======================================================================= PrnV
*/
void PrnV( VECTOR a )
  {
  vMsg("%12.5e %12.5e %12.5e\n", a.x, a.y, a.z);
  }

/*======================================================================= PrnM
*/
void PrnM( MATRIX a )
  {
  vMsg("%12.5e %12.5e %12.5e\n", a.xx, a.xy, a.xz);
  vMsg("%12.5e %12.5e %12.5e\n", a.yx, a.yy, a.yz);
  vMsg("%12.5e %12.5e %12.5e\n", a.zx, a.zy, a.zz);
  }

/*======================================================================= MulSV
*/
VECTOR MulSV( float f, VECTOR a )
  {
  VECTOR b;
  b.x = f*a.x;
  b.y = f*a.y;
  b.z = f*a.z;
  return b;
  }

/*======================================================================= MulSM
*/
MATRIX MulSM( float f, MATRIX a )
  {
  MATRIX b;
  float *pa, *pb;
  int i;
  pa = &a.xx;
  pb = &b.xx;
  for (i=0; i<9; i++)  *pb++ = f*(*pa++);
  return b;
  }

/*======================================================================= DotVV
*/
float DotVV( VECTOR a, VECTOR b )
  {
  float s;
  s = a.x*b.x + a.y*b.y + a.z*b.z;
  return s;
  }

/*======================================================================== AbsV
*/
float AbsV( VECTOR a )
  {
  float s;
  s = sqrt(a.x*a.x + a.y*a.y + a.z*a.z);
  return s;
  }

/*======================================================================= CroVV
*/
VECTOR CroVV( VECTOR a, VECTOR b )
  {
  VECTOR r;
  r.x = a.y*b.z - a.z*b.y;
  r.y = a.z*b.x - a.x*b.z;
  r.z = a.x*b.y - a.y*b.x;
  return r;
  }

/*======================================================================= DotVM
*/
VECTOR DotVM( VECTOR a, MATRIX b )
  {
  VECTOR r;
  r.x = a.x*b.xx + a.y*b.yx + a.z*b.zx;
  r.y = a.x*b.xy + a.y*b.yy + a.z*b.zy;
  r.z = a.x*b.xz + a.y*b.yz + a.z*b.zz;
  return r;
  }

/*====================================================================== DotMV
*/
VECTOR DotMV( MATRIX a, VECTOR b )
  {
  VECTOR r;
  r.x = a.xx*b.x + a.xy*b.y + a.xz*b.z;
  r.y = a.yx*b.x + a.yy*b.y + a.yz*b.z;
  r.z = a.zx*b.x + a.zy*b.y + a.zz*b.z;
  return r;
  }

/*====================================================================== DotMM
*/
MATRIX DotMM( MATRIX a, MATRIX b )
  {
  MATRIX r;
  r.xx = a.xx*b.xx + a.xy*b.yx + a.xz*b.zx;
  r.xy = a.xx*b.xy + a.xy*b.yy + a.xz*b.zy;
  r.xz = a.xx*b.xz + a.xy*b.yz + a.xz*b.zz;
  r.yx = a.yx*b.xx + a.yy*b.yx + a.yz*b.zx;
  r.yy = a.yx*b.xy + a.yy*b.yy + a.yz*b.zy;
  r.yz = a.yx*b.xz + a.yy*b.yz + a.yz*b.zz;
  r.zx = a.zx*b.xx + a.zy*b.yx + a.zz*b.zx;
  r.zy = a.zx*b.xy + a.zy*b.yy + a.zz*b.zy;
  r.zz = a.zx*b.xz + a.zy*b.yz + a.zz*b.zz;
  return r;
  }

/*======================================================================= TrnM
*/
MATRIX TrnM( MATRIX a )
  {
  MATRIX r;
  r.xx = a.xx;  r.xy = a.yx;  r.xz = a.zx;
  r.yx = a.xy;  r.yy = a.yy;  r.yz = a.zy;
  r.zx = a.xz;  r.zy = a.yz;  r.zz = a.zz;
  return r;
  }

/*======================================================================== DefV
*/
VECTOR DefV( float x, float y, float z )
  {
  VECTOR r;
  r.x = x;  r.y = y;  r.z = z;
  return r;
  }

/*======================================================================== DefM
*/
MATRIX DefM(
float xx, float xy, float xz,
float yx, float yy, float yz,
float zx, float zy, float zz )
  {
  MATRIX r;
  r.xx = xx;  r.xy = xy;  r.xz = xz;
  r.yx = yx;  r.yy = yy;  r.yz = yz;
  r.zx = zx;  r.zy = zy;  r.zz = zz;
  return r;
  }

/*======================================================================= AddVV
*/
VECTOR AddVV( VECTOR a, VECTOR b )
  {
  VECTOR r;
  r.x = a.x + b.x;
  r.y = a.y + b.y;
  r.z = a.z + b.z;
  return r;
  }

/*======================================================================= SubVV
*/
VECTOR SubVV( VECTOR a, VECTOR b )
  {
  VECTOR r;
  r.x = a.x - b.x;
  r.y = a.y - b.y;
  r.z = a.z - b.z;
  return r;
  }

/*======================================================================= AddMM
*/
MATRIX AddMM( MATRIX a, MATRIX b )
  {
  MATRIX r;
  float *pa, *pb, *pr;
  int i;
  pa = &a.xx;
  pb = &b.xx;
  pr = &r.xx;
  for (i=0; i<9; i++)  *pr++ = *pa++ + *pb++;
  return r;
  }

/*======================================================================= SubMM
*/
MATRIX SubMM( MATRIX a, MATRIX b )
  {
  MATRIX r;
  float *pa, *pb, *pr;
  int i;
  pa = &a.xx;
  pb = &b.xx;
  pr = &r.xx;
  for (i=0; i<9; i++)  *pr++ = *pa++ - *pb++;
  return r;
  }

/*======================================================================= DyaVV
*/
MATRIX DyaVV( VECTOR a, VECTOR b )
  {
  MATRIX r;
  r.xx = a.x*b.x;  r.xy = a.x*b.y;  r.xz = a.x*b.z;
  r.yx = a.y*b.x;  r.yy = a.y*b.y;  r.yz = a.y*b.z;
  r.zx = a.z*b.x;  r.zy = a.z*b.y;  r.zz = a.z*b.z;
  return r;
  }

/*======================================================================= DefR2
*/
MATRIX DefR2( float x, float y )
  {
  MATRIX r;
  float a;
  a = sqrt(x*x + y*y);
  if (a == 0.0) { x = 1;  a = 1; }
  r.xx = x/a;    r.yx = y/a;   r.zx = 0;
  r.xy = -r.yx;  r.yy = r.xx;  r.zy = 0;
  r.xz = 0;      r.yz = 0;     r.zz = 1;
  return r;
  }

/*======================================================================== ChlM
*/
MATRIX ChlM( MATRIX a )
  {
  dP(ChlM);
  float x, *pa, *pr;
  MATRIX r;
  int i, j, k;
  pa = &a.xx;
  pr = &r.xx;
  for (i=0; i<3; i++)
    for (j=i; j<3; j++) {
      x = pa[3*i+j];
      for (k=i-1; k>=0; k--)  x -= pr[3*j+k]*pr[3*i+k];
      if (i==j) {
        if (x < 0.0)                            eX(1);
        pr[3*i+i] = sqrt(x);
        }
      else {
        if (pr[3*i+i] == 0.0) {
          if (x != 0.0)                         eX(2);
          pr[3*j+i] = 0.0;
          }
        else  pr[3*j+i] = x/pr[3*i+i];
        pr[3*i+j] = 0.0; }
      }
  return r;
eX_1:  eX_2:
  PrnM(a);
  vMsg("*** cholesky separation not possible! (%s)\n", ePGMs);
  exit(1);
  }

/*======================================================================= DetM
*/
float DetM( MATRIX a )
  {
  float r;
  r = a.xx*a.yy*a.zz + a.xy*a.yz*a.zx + a.xz*a.yx*a.zy
    - a.xz*a.yy*a.zx - a.xx*a.yz*a.zy - a.xy*a.yx*a.zz;
  return r;
  }

/*======================================================================= InvM
*/
MATRIX InvM( MATRIX a )
  {
  dP(InvM);
  MATRIX r;
  float d;
  d = DetM(a);  if (d == 0.0)           eX(1);
  r.xx =  (a.yy*a.zz - a.yz*a.zy)/d;
  r.xy = -(a.xy*a.zz - a.xz*a.zy)/d;
  r.xz =  (a.xy*a.yz - a.xz*a.yy)/d;
  r.yx = -(a.yx*a.zz - a.yz*a.zx)/d;
  r.yy =  (a.xx*a.zz - a.xz*a.zx)/d;
  r.yz = -(a.xx*a.yz - a.xz*a.yx)/d;
  r.zx =  (a.yx*a.zy - a.yy*a.zx)/d;
  r.zy = -(a.xx*a.zy - a.xy*a.zx)/d;
  r.zz =  (a.xx*a.yy - a.xy*a.yx)/d;
  return r;
eX_1:
  PrnM(a);
  vMsg("*** singular matrix! (%s)\n", ePGMs);
  exit(1);
  }

/*======================================================================== GssJrd
 *
 *  Solution of systems of linear equations by Gauss-Jordan elimination.
 *  Adapted from Numerical Recipes.
 *
 *  The matrix amat is replaced by its inverse,
 *  the right hand sides bmat are replaced by the solution vectors.
 *
 ==============================================================================*/
#define SWAP(a,b) {float temp=(a);(a)=(b);(b)=temp;}
#define a(i,j)  amat[((i)-1)*n+(j)-1]
#define b(i,j)  bmat[((i)-1)*m+(j)-1]

int GssJrd(     /* Gauss-Jordan elimination */
int n,          /* number of unknowns       */
float *amat,    /* n x n matrix elements    */
int m,          /* number of rhs-vectors    */
float *bmat)    /* n x m matrix elements    */
    {
    dP(GssJrd);
    int *indxc, *indxr, *ipiv;
    int i, icol, irow, j, k, l, ll;
    float big, dum, pivinv;
    indxc = ALLOC((n+1)*sizeof(int));  if (!indxc)      eX(1);
    indxr = ALLOC((n+1)*sizeof(int));  if (!indxr)      eX(2);
    ipiv  = ALLOC((n+1)*sizeof(int));  if (!ipiv)       eX(3);
    for (j=1; j<=n; j++) ipiv[j] = 0;
    for (i=1; i<=n; i++) {
        big = 0.0;
        for (j=1; j<=n; j++)
            if (ipiv[j] != 1)
                for (k=1; k<=n; k++) {
                    if (ipiv[k] == 0) {
                        if (fabs(a(j,k)) >= big) {
                            big = fabs(a(j,k));
                            irow = j;
                            icol = k;
                            }
                        } else if (ipiv[k] > 1)         eX(4);
                    }
        ++(ipiv[icol]);
        if (irow != icol) {
            for (l=1; l<=n; l++)  SWAP(a(irow,l), a(icol,l))
            for (l=1; l<=m; l++)  SWAP(b(irow,l), b(icol,l))
            }
        indxr[i] = irow;
        indxc[i] = icol;
        if (a(icol,icol) == 0.0)                        eX(5);
        pivinv = 1.0/a(icol,icol);
        a(icol,icol) = 1.0;
        for (l=1; l<=n; l++)  a(icol,l) *= pivinv;
        for (l=1; l<=m; l++)  b(icol,l) *= pivinv;
        for (ll=1; ll<=n; ll++)
            if (ll != icol) {
                dum = a(ll,icol);
                a(ll,icol) = 0.0;
                for (l=1; l<=n; l++) a(ll,l) -= a(icol,l)*dum;
                for (l=1; l<=m; l++) b(ll,l) -= b(icol,l)*dum;
                }
        }
    for (l=n; l>=1; l--) {
        if (indxr[l] != indxc[l])
            for (k=1; k<=n; k++)  SWAP(a(k,indxr[l]),a(k,indxc[l]));
        }
    FREE(ipiv);
    FREE(indxr);
    FREE(indxc);
    return 0;
eX_1: eX_2: eX_3:
  eMSG("%d integers not allocated! (%s)\n", n+1, ePGMs);
eX_4: eX_5:
  FREE(ipiv);
  FREE(indxr);
  FREE(indxc);
  eMSG("singular matrix! (%s)\n", ePGMs);
  }
#undef SWAP
#undef a
#undef b

#ifdef  MAIN  /*############################################################*/

float A, B, C, D;
VECTOR X, Y, Z, V0;
MATRIX U, V, W, M0;

enum CMD {
    NOCMD,   PRNS,   PRNV,   PRNM,   DEFV,  DEFM,  DEFR2,  DYAVV,  MULSV,
    MULSM,  DOTVV,  DOTVM,  DOTMV,  DOTMM,  TRNM,   INVM,   CHLM,   DETM,
    ADDVV,  SUBVV,  ADDMM,  SUBMM,   ERFC };
char *sCmd[] = {
       "",  "prns", "prnv","prnm", "defv", "defm","defr2","dyavv","mulsv",
   "mulsm","dotvv","dotvm","dotmv","dotmm","trnm", "invm", "chlm", "detm",
   "addvv","subvv","addmm","submm", "erfc", NULL };

enum CMD GetCmd( char *cmd )
  {
  int i;
  for (i=0; (sCmd[i]); i++)  if (!strcmp(cmd, sCmd[i]))  break;
  if (!sCmd[i])  return 0;
  return i;
  }

float GetS( char *name )
  {
  if (!strcmp(name, "a"))  return A;
  if (!strcmp(name, "b"))  return B;
  if (!strcmp(name, "c"))  return C;
  printf("unknown scalar %s!\n", name);
  return 0.0;
  }

VECTOR GetV( char *name )
  {
  if (!strcmp(name, "x"))  return X;
  if (!strcmp(name, "y"))  return Y;
  if (!strcmp(name, "z"))  return Z;
  printf("unknown vector %s!\n", name);
  return V0;
  }

MATRIX GetM( char *name )
  {
  if (!strcmp(name, "u"))  return U;
  if (!strcmp(name, "v"))  return V;
  if (!strcmp(name, "w"))  return W;
  printf("unknown matrix %s!\n", name);
  return M0;
  }

void SetS( char *name, float r )
  {
  if (!strcmp(name, "a"))  A = r;
  if (!strcmp(name, "b"))  B = r;
  if (!strcmp(name, "c"))  C = r;
  }

void SetV( char *name, VECTOR r )
  {
  if (!strcmp(name, "x"))  X = r;
  if (!strcmp(name, "y"))  Y = r;
  if (!strcmp(name, "z"))  Z = r;
  }

void SetM( char *name, MATRIX r )
  {
  if (!strcmp(name, "u"))  U = r;
  if (!strcmp(name, "v"))  V = r;
  if (!strcmp(name, "w"))  W = r;
  }

int main()
  {
  char s[80], *pc, *pn, rn[80], cs[80], opa[80], opb[80];
  char *sep = " ,;\t\r";
  static float f, x, y, z, xx, xy, xz, yx, yy, yz, zx, zy, zz;
  int n;
  enum CMD cmd;
  printf("Test von IBJmath\n");
  for (;;) {
    printf(": ");  fflush(stdout);
    gets(s);
    if (!*s)  break;
    pn = s;
    pc = MsgTok(pn, &pn, sep, &n);
    if (!pc)  continue;
    *rn = 0;
    strncat(rn, pc, n);
    pc = MsgTok(pn, &pn, sep, &n);
    if (!pc)  continue;
    *cs = 0;
    strncat(cs, pc, n);
    cmd = GetCmd(cs);
    if (!cmd) {
      printf("unknown command %s!\n", cs);
      continue;
      }
    pc = MsgTok(pn, &pn, sep, &n);
    if (!pc)  continue;
    *opa = 0;
    strncat(opa, pc, n);
    switch (cmd) {
      case PRNS:  printf("%12.5e\n", GetS(opa));
                  break;
      case PRNV:  PrnV(GetV(opa));
                  break;
      case PRNM:  PrnM(GetM(opa));
                  break;
      case DEFV:  sscanf(pc, "%f%f%f", &x, &y, &z);
                  SetV(rn, DefV(x, y, z));
                  break;
      case DEFM:  sscanf(pc, "%f%f%f%f%f%f%f%f%f",
                     &xx, &xy, &xz, &yx, &yy, &yz, &zx, &zy, &zz);
                  SetM(rn, DefM(xx, xy, xz, yx, yy, yz, zx, zy, zz));
                  break;
      case DEFR2: sscanf(pc, "%f%f", &x, &y);
                  SetM(rn, DefR2(x, y));
                  break;
      case DYAVV: sscanf(pn, "%s", opb);
                  SetM(rn, DyaVV(GetV(opa), GetV(opb)));
                  break;
      case MULSV: sscanf(pc, "%f%s", &f, opb);
                  SetV(rn, MulSV(f, GetV(opb)));
                  break;
      case MULSM: sscanf(pc, "%f%s", &f, opb);
                  SetM(rn, MulSM(f, GetM(opb)));
                  break;
      case DOTVV: sscanf(pn, "%s", opb);
                  SetS(rn, DotVV(GetV(opa), GetV(opb)));
                  break;
      case DOTVM: sscanf(pn, "%s", opb);
                  SetV(rn, DotVM(GetV(opa), GetM(opb)));
                  break;
      case DOTMV: sscanf(pn, "%s", opb);
                  SetV(rn, DotMV(GetM(opa), GetV(opb)));
                  break;
      case DOTMM: sscanf(pn, "%s", opb);
                  SetM(rn, DotMM(GetM(opa), GetM(opb)));
                  break;
      case TRNM:  SetM(rn, TrnM(GetM(opa)));
                  break;
      case INVM:  SetM(rn, InvM(GetM(opa)));
                  break;
      case CHLM:  SetM(rn, ChlM(GetM(opa)));
                  break;
      case DETM:  SetS(rn, DetM(GetM(opa)));
                  break;
      case ADDVV: sscanf(pn, "%s", opb);
                  SetV(rn, AddVV(GetV(opa), GetV(opb)));
                  break;
      case SUBVV: sscanf(pn, "%s", opb);
                  SetV(rn, SubVV(GetV(opa), GetV(opb)));
                  break;
      case ADDMM: sscanf(pn, "%s", opb);
                  SetM(rn, AddMM(GetM(opa), GetM(opb)));
                  break;
      case SUBMM: sscanf(pn, "%s", opb);
                  SetM(rn, SubMM(GetM(opa), GetM(opb)));
                  break;
      case ERFC:  sscanf(pc, "%f", &f);
                  SetS(rn, Erfc(f));
                  break;
      default:    printf("unknown command %s (%d)!\n", cs, cmd);
      }
    }
  printf("finished\n");
  return 0;
  }
#endif  /*###################################################################*/
/*===========================================================================*/
